// sound
// subsim (C)+(W) Markus Petermann. SEE LICENSE

#include <SDL.h>

#include <iostream>
#include <fstream>
#include <math.h>
#include <string>
using namespace std;
#include "system.h"
#include "sound.h"



//
// Class sound
//
bool sound::initialized = false;
bool sound::use_sound = true;

SDL_AudioSpec* sound::audio_spec_ptr = 0;
SDL_AudioSpec sound::audio_spec;
struct sound::sample sound::sounds[sound::number_of_samples];

void sound::mix_audio ( void* unused, Uint8* stream, int len )
{
	if (!use_sound) return;

	for ( int i = 0; i < number_of_samples; i++ )
	{
		Uint32 amount = sounds[i].dlen - sounds[i].dpos;

		if ( amount > len )
			amount = len;

		SDL_MixAudio ( stream, &sounds[i].data[sounds[i].dpos], amount,
			sounds[i].dvolume );

		sounds[i].dpos += amount;
	}
}

sound::sound ()
{
	init ();
}

sound::sound ( const string& filename )
{
	init ();
	load_sound ( filename );
}

sound::~sound ()
{
	if ( sound_data )
		free ( sound_data );
}

void sound::init ()
{
	sound_data = 0;
	sound_data_len = 0;
	
	open_audio_device ();	// initialize sound if not done yet
}

void sound::open_audio_device ()
{
	if (use_sound && !initialized) {
		int errval = SDL_InitSubSystem(SDL_INIT_AUDIO);
		if (errval == -1) {
			string s = "SDL: sound initialization failed. SDL error: "
				+ string(SDL_GetError());
			system::sys().add_console(s);
			return;
		}

		SDL_AudioSpec desired;
		desired.freq     = 22050;
		desired.format   = AUDIO_S16;
		desired.channels = 2;
		desired.samples  = 512;
		desired.callback = mix_audio;
		desired.userdata = NULL;

		if ( SDL_OpenAudio ( &desired, &audio_spec ) == -1 )
		{
			system::sys().add_console ( "Failed to open audio device!" );
		}
		audio_spec_ptr = &audio_spec;
		memset ( sounds, 0x00, sizeof sounds );
		SDL_PauseAudio ( 0 );
		
		initialized = true;
	}
}

void sound::close_audio_device ()
{
	if (initialized)
		SDL_CloseAudio ();
}

void sound::play ( double volume )
{
	if (!use_sound) return;

	int index = 0;

	if ( !sound_data || !sound_data_len )
		return;

	// Find an empty sample first. The following condition is only met when
	// dlen = dpos = 0, i.e. the sound is not used, or dlen = dpos, i.e.
	// the sound is played and finished.
	for ( index = 0; index < number_of_samples; index++ )
	{
		if ( sounds[index].dlen == sounds[index].dpos )
			break;
	}
	
	if (index == number_of_samples) {
		system::sys().add_console("failed to get empty sample slot!");
		return;
	}

	if ( volume < 0.0f )
		volume = 0.0f;
	if ( volume > 1.0f )
		volume = 1.0f;

	// Lock callback function. This prevents SDL from calling the mix_audio
	// callback.
	SDL_LockAudio ();
	sounds[index].data = sound_data;
	sounds[index].dlen = sound_data_len;
	sounds[index].dpos = 0;
	sounds[index].dvolume = int ( SDL_MIX_MAXVOLUME * volume );
	// Unlock callback function.
	SDL_UnlockAudio ();
}

void sound::load_sound ( const string& filename )
{
	if (!use_sound) return;

	SDL_AudioSpec wave;
	Uint8* data = 0;
	Uint32 data_len = 0;
	SDL_AudioCVT cvt;

	// Load audio file.
	if ( SDL_LoadWAV ( filename.c_str (), &wave, &data, &data_len ) == NULL )
	{
		system::sys().add_console ( ( string ( "Failed to load audio file " ) + filename ) );
		return;
	}

	// Now convert data.
	// Build converter.
	if ( SDL_BuildAudioCVT ( &cvt, wave.format, wave.channels, wave.freq,
			audio_spec_ptr->format, audio_spec_ptr->channels,
			audio_spec_ptr->freq ) < 0 )
	{
		system::sys().myassert ( 0, "Cannot build converter" );
	}

	// Prepare conversion of sound datas.
	cvt.buf = (Uint8*) malloc ( data_len * cvt.len_mult );
	system::sys().myassert ( cvt.buf != 0, "Failed to allocate memory for cvt.buf!" );
	memcpy ( cvt.buf, data, data_len );
	cvt.len = data_len;
	SDL_ConvertAudio ( &cvt );

	// Now store datas.
	sound_data = cvt.buf;
	sound_data_len = cvt.len_cvt;

	SDL_FreeWAV ( data );
}
